"""This file contains client-side utilities provided to make it easier to use Specklia."""
from datetime import datetime
import os
from typing import Dict, List, Optional

import geopandas as gpd
import numpy as np
import rasterio
from shapely.geometry import shape


def save_gdf_as_tiff(
        gdf: gpd.GeoDataFrame, data_col: str, bounds: Dict[str, float],
        output_path: str, xy_proj4: Optional[str] = None, data_type: str = 'float32') -> None:
    """
    Save a GeoDataFrame as a GeoTIFF file.

    This function is provided so that raster data returned from Specklia in the form of GeoDataFrames
    can be easily converted back to GeoTIFF for integration with other tools.

    Note that this function intentionally *does not* check whether the supplied GeoDataFrame will result
    in a full raster after transformation. Instead, a grid is defined based on the bounds arguments
    and the sorted unique values in gdf, and the data_col values directly assigned to it.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        The GeoDataFrame to save to a GeoTIFF file.
    data_col : str
        The name of the column within the GeoDataFrame to save out as a Tiff file.
    bounds : Dict[str, float]:
        A dictionary containing the keys "min_x", "min_y", "max_x" and "max_y" indicating the bounds of the saved tiff.
        These are provided separately because the data in gdf may not extend to the desired edges of the tiff file.
    output_path : str
        The output path of the GeoTIFF file.
    xy_proj4 : Optional[str], by default None
        If not None, the Proj4 code of the 'x' and 'y' columns in the GeoDataFrame. These columns will then be used
        to generate the raster instead of the GeoDataFrame's geometry.
        If None, the GeoDataFrame's geometry is used to generate the raster instead.
    data_type : str
        The data type to save data_col as, by default 'float32'
    """
    # we start by working out the desired axes of the output raster
    if xy_proj4 is not None:
        # use the 'x' and 'y' columns in the GeoDataFrame
        x_col = gdf['x']
        y_col = gdf['y']
        crs = xy_proj4
    else:
        # we use the geometry within the GeoDataFrame.
        x_col = gdf.geometry.x
        y_col = gdf.geometry.y
        crs = gdf.geometry.crs

    # TODO: The below lines will only work if there's at least one pair of adjacent pixels in X
    # and at least one pair of adjacent pixels in Y. We'll reload Specklia's raster data at a later
    # date to explicitly store the x and y resolutions to mitigate this, but it's a low priority problem.
    dx = np.min(np.diff(np.sort(x_col.unique())))
    dy = np.min(np.diff(np.sort(y_col.unique())))

    # generate all of the points we want to end up with in the output raster
    # we need to offset both these axes by one in order to use np.searchsorted() in a manner that matches
    # how the EOLIS Gridded products were loaded into Specklia.
    desired_x_axis = np.arange(bounds['min_x'], bounds['max_x'], dx) + dx
    desired_y_axis = np.arange(bounds['min_y'], bounds['max_y'], dy) + dy

    # create the output raster, but fill it with NaN
    gridded_data = np.full((len(desired_y_axis), len(desired_x_axis)), np.nan)

    # set the valid points within it
    gridded_data[np.searchsorted(desired_y_axis, y_col),
                 np.searchsorted(desired_x_axis, x_col)] = gdf[data_col]

    # finally, save the raster to file.
    # There's a lot of wierdness here w.r.t axes orientation that we have to replicate
    # in order to maintain compatibility with the Timeseries Service.
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with rasterio.open(
            output_path,
            'w',
            driver='GTiff',
            height=gridded_data.shape[0],
            width=gridded_data.shape[1],
            count=1,
            dtype=data_type,
            crs=crs,
            transform=rasterio.transform.from_origin(bounds['min_x'], bounds['max_y'], dx, dy),
            compress='lzw',
            nodata=np.nan) as rst:
        rst.write_band(1, np.flipud(gridded_data))


def deserialise_sources(sources: List[Dict]) -> List[Dict]:
    """
    Reverse some serialisation of sources returned from /query.

    Reverses some serialisation of the sources dictionary returned from the /query endpoint for end-user convenience.
    Convert the WKB coverage polygon into a Shapely geometry object, and the min and max times into datetimes.

    Parameters
    ----------
    sources: List[Dict]
        A list of sources returned from Specklia

    Returns
    -------
    List[Dict]
        Sources after the coverage polygon, min_time and max_time have been deserialised.
    """
    for source in sources:
        source['geospatial_coverage'] = shape(source['geospatial_coverage'])
        source['min_time'] = datetime.fromisoformat(source['min_time'])
        source['max_time'] = datetime.fromisoformat(source['max_time'])

    return sources



"""Unit tests for chunked_transfer.py."""
from http import HTTPStatus
import struct
from unittest.mock import call, MagicMock, patch

import pandas as pd

from specklia import chunked_transfer


def test_split_into_chunks():
    assert chunked_transfer.split_into_chunks(b'abcdefghijklmnop', chunk_size=5) == [
        (1, b'abcde'), (2, b'fghij'), (3, b'klmno'), (4, b'p')]


def test_merge_from_chunks():
    assert chunked_transfer.merge_from_chunks([
        (1, b'abcde'), (2, b'fghij'), (3, b'klmno'), (4, b'p')]) == b'abcdefghijklmnop'


def test_upload_chunks():
    with patch('specklia.chunked_transfer.requests.post') as mock_post:
        mock_post.return_value.status_code = HTTPStatus.OK
        mock_post.return_value.json.return_value = {'chunk_set_uuid': 'cheese'}

        assert chunked_transfer.upload_chunks(
            api_address='wibble', chunks=[(1, b'a'), (2, b'b')], logger=MagicMock(name="mock_logger")) == 'cheese'

        mock_post.assert_has_calls([
            call('wibble/chunk/upload/1-of-2', data=b'a'),
            call().json(),
            call().json(),
            call('wibble/chunk/upload/cheese/2-of-2', data=b'b')])


def test_download_chunks():
    with patch('specklia.chunked_transfer.requests.get') as mock_get:
        mock_get.side_effect = [
            MagicMock(name="mock_response_1", status_code=HTTPStatus.OK, content=struct.pack('i', 1) + b'wibble'),
            MagicMock(name="mock_response_2", status_code=HTTPStatus.OK, content=struct.pack('i', 2) + b'wobble'),
            MagicMock(name="mock_response_3", status_code=HTTPStatus.NO_CONTENT, content=b'')]

        assert chunked_transfer.download_chunks(api_address='wibble', chunk_set_uuid='rawr') == [
            (1, b'wibble'), (2, b'wobble')]


def test_serialise_dataframe_roundtrip():
    df = pd.DataFrame({'a': [1, 1, 2, 3], 'b': ['alfred', 'dave', 'ken', 'sally'], 'c': [1, 2, 4, 4.4]})

    pd.testing.assert_frame_equal(
        df, chunked_transfer.deserialise_dataframe(chunked_transfer.serialise_dataframe(df)))



"""
Interface to Mongo for using Mongo as a buffer for chunked data transfer.

We use Mongo as a buffer because we cannot guarantee that all of the requests
for individual chunks will hit the same worker. While we could use streamed responses for the download,
they're not available for upload, so for simplicity we use the same approach in both directions.

The intended usage pattern is that a single message is stored as a single "chunk set".
The chunk set is first "filled" (either by the client or the server), then "emptied" to obtain the data
(again, by either the client or the server).

Note that while this can be used for pagination, it is not in itself pagination.

We plan to gather most of this material into ew_common after the chunked transfer interface has been rolled out
to its three main users (ew_geostore, ew_specklia, ew_online_processing_service) and proven effective for each.
At that point, this entire module will move into ew_common. Note that the chunked transfer interface will always
require MongoDB or a similar provision to work correctly.

IMPORTANT: THE VERSION HERE IN THE SPECKLIA PACKAGE MUST NOT BE MADE DEPENDENT UPON EW_COMMON SINCE EW_COMMON
IS PRIVATE BUT THIS PACKAGE IS PUBLIC!
"""

from enum import Enum
from http import HTTPStatus
from io import BytesIO
from logging import Logger
import struct
from typing import List, Tuple, Union

from geopandas import GeoDataFrame, read_feather as read_geofeather
from pandas import DataFrame, read_feather
import requests

CHUNK_DB_NAME = "data_transfer_chunks"
CHUNK_METADATA_COLLECTION_NAME = "chunk_metadata"
MAX_CHUNK_AGE_SECONDS = 3600
MAX_CHUNK_SIZE_BYTES = 5 * 1024 ** 2  # must be small enough to fit into an HTTP GET Request


class ChunkSetStatus(Enum):
    """
    Chunk set status.

    Prevents the accidental access of chunk sets that have not yet received all of their data.
    """

    FILLING = 0
    EMPTYING = 1


def upload_chunks(api_address: str, chunks: List[Tuple[int, bytes]], logger: Logger) -> str:
    """
    Upload data chunks.

    Upload a series of data chunks through the chunked transfer mechanism.
    This method is for use on the client, not the server.

    Parameters
    ----------
    api_address : str
        The full URL of the API, including port but not including endpoint, e.g. "http://127.0.0.1:9999"
    chunks : List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk
    logger : Logger
        A logger with which to log the upload.

    Returns
    -------
    str
        The chunk set uuid of the uploaded chunks
    """
    # post the first chunk to start the upload
    response = requests.post(
        api_address + f"/chunk/upload/{chunks[0][0]}-of-{len(chunks)}",
        data=chunks[0][1])
    logger.info("response from very first /chunk/upload was '%s'", response.json())
    assert response.status_code == HTTPStatus.OK, response.text
    chunk_set_uuid = response.json()['chunk_set_uuid']

    # post the rest of the chunks in a random order
    for i, chunk in chunks[1:]:
        response = requests.post(
            api_address + f"/chunk/upload/{chunk_set_uuid}/{i}-of-{len(chunks)}", data=chunk)
        logger.info("response from subsequent /chunk/upload/uuid call was '%s'", response.text)
        assert response.status_code == HTTPStatus.OK, response.text

    return chunk_set_uuid


def download_chunks(api_address: str, chunk_set_uuid: str) -> List[Tuple[int, bytes]]:
    """
    Download data chunks.

    Download a series of data chunks through the chunked transfer mechanism.
    This method is for use on the client, not the server.

    Parameters
    ----------
    api_address : str
        The full URL of the API, including port but not including endpoint, e.g. "http://127.0.0.1:9999"
    chunk_set_uuid : str
        The uuid of the chunk set to download.

    Returns
    -------
    chunks : List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk
    """
    # fetch the data
    data_chunks = []
    finished = False
    while not finished:
        this_chunk_response = requests.get(api_address + f"/chunk/download/{chunk_set_uuid}")
        if this_chunk_response.status_code == HTTPStatus.NO_CONTENT:
            finished = True
        else:
            data_chunks.append((
                struct.unpack('i', this_chunk_response.content[:4])[0],
                this_chunk_response.content[4:]))

    return data_chunks


def split_into_chunks(data: bytes, chunk_size: int = MAX_CHUNK_SIZE_BYTES) -> List[Tuple[int, bytes]]:
    """
    Split data into compressed chunks for transport.

    Parameters
    ----------
    data : bytes
        The data to be split into chunks.
    chunk_size: int
        The maximum number of bytes allowed in each chunk.

    Returns
    -------
    List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk
    """
    return list(
        enumerate((data[i:i + chunk_size] for i in range(0, len(data), chunk_size)), start=1))


def merge_from_chunks(chunks: List[Tuple[int, bytes]]) -> bytes:
    """
    Merge data that has been split into compressed chunks back into a single message.

    Parameters
    ----------
    chunks : List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk

    Returns
    -------
    bytes
        The merged data
    """
    return b''.join([dc[1] for dc in sorted(chunks, key=lambda x: x[0])])


def deserialise_dataframe(data: bytes) -> Union[DataFrame, GeoDataFrame]:
    """
    Convert a binary serialised feather table to pandas dataframe.

    Parameters
    ----------
    data : bytes
        Binary serialised feather table.

    Returns
    -------
    Union[DataFrame, GeoDataFrame]
        Input table converted to a pandas dataframe.

    Raises
    ------
    ValueError
        When bytes can't be interpreted as meaningful dataframe.
    """
    try:
        buffer = BytesIO(data)
        df = read_geofeather(buffer)
    except ValueError as e:
        # First attempt to deserialise as a geodataframe. If geo meta is missing, we expect a clear ValueError
        # and we then load as a plain dataframe instead.
        if "Missing geo meta" in e.args[0] or "'geo' metadata" in e.args[0]:
            try:
                df = read_feather(BytesIO(data))
            except ValueError as e:
                raise ValueError("Couldn't deserialise table format") from e
        else:
            raise ValueError("Couldn't deserialise table format") from e
    return df


def serialise_dataframe(df: Union[DataFrame, GeoDataFrame]) -> bytes:
    """
    Serialise a dataframe using the feather table format.

    Parameters
    ----------
    df : DataFrame
        Input dataframe

    Returns
    -------
    bytes
        Serialised feather table.
    """
    feather_buffer = BytesIO()
    # Browser implementations of feather do not support compressed feather formats.
    df.to_feather(feather_buffer, compression='uncompressed')
    feather_buffer.seek(0)
    return feather_buffer.getvalue()
